/*
 * Copyright (C) 2010 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Mikkel Kamstrup Erlandsen <mikkel.kamstrup@canonical.com>
 *
 */
using Dee;
using DBus;
using Zeitgeist;
using Zeitgeist.Timestamp;
using Unity.Place;
using Config;
using Gee;

namespace Unity.FilesPlace {

  public class Daemon : GLib.Object
  {
    private Zeitgeist.Log log;
    private Zeitgeist.Index index;
    private Zeitgeist.Monitor monitor;

    private Unity.Place.Controller control;
    private Unity.Place.EntryInfo files;

    /* For each section we have a set of Zeitgeist.Event templates that
     * we use to query Zeitgeist */
    private Gee.List<PtrArray> section_templates;

    /* Store a maping of DateMonth to Dee.ModelIter. We map to the iter and
     * not simply the offset, because in theory anyone on the bus could
     * update the Dee.SharedModel we use for the groups changing the row
     * offsets*/
    private Gee.List<unowned Dee.ModelIter?> months;

    construct
    {
      var sections_model = new Dee.SharedModel(
                                 "com.canonical.Unity.FilesPlace.SectionsModel",
                                 2, typeof (string), typeof (string));

      var groups_model = new Dee.SharedModel(
                                 "com.canonical.Unity.FilesPlace.GroupsModel",
                                 3, typeof (string), typeof (string),
                                 typeof (string));

      var global_groups_model = new Dee.SharedModel(
                                 "com.canonical.Unity.FilesPlace.GlobalGroupsModel",
                                 3, typeof (string), typeof (string),
                                 typeof (string));

      var results_model = new Dee.SharedModel(
                                 "com.canonical.Unity.FilesPlace.ResultsModel",
                                 6, typeof (string), typeof (string),
                                 typeof (uint), typeof (string),
                                 typeof (string), typeof (string));

      var global_results_model = new Dee.SharedModel(
                                 "com.canonical.Unity.FilesPlace.GlobalResultsModel",
                                 6, typeof (string), typeof (string),
                                 typeof (uint), typeof (string),
                                 typeof (string), typeof (string));

      section_templates = new Gee.ArrayList<PtrArray> ();
      prepare_section_templates();

      files = new EntryInfo ("/com/canonical/unity/filesplace/files");
      files.sections_model = sections_model;
      files.entry_renderer_info.groups_model = groups_model;
      files.entry_renderer_info.results_model = results_model;
      files.global_renderer_info.groups_model = global_groups_model;
      files.global_renderer_info.results_model = global_results_model;

      populate_sections ();
      populate_groups ();

      files.icon = @"$(Config.PREFIX)/share/unity/files.png";


      // FIXME: We monitor on all events, restrict templates to file events
      var templates = new PtrArray();
      var event = new Zeitgeist.Event ();
      templates.add (event);
      monitor = new Zeitgeist.Monitor (new Zeitgeist.TimeRange.from_now (),
                                       (owned) templates);

      log = new Zeitgeist.Log();
      index = new Zeitgeist.Index();

      /* Listen for section changes */
      files.notify["active-section"].connect (
        (obj, pspec) => {
          var section = (Section)files.active_section;
          var _results_model = files.entry_renderer_info.results_model;
          var _groups_model = files.entry_renderer_info.groups_model;
          if (search_is_invalid (files.active_search))
            {
              update_without_search_async.begin(section,
                                                _results_model, _groups_model);
            }
          else
            {
              update_search_async.begin (files.active_search, section,
                                         _results_model, _groups_model);
            }
        }
      );

      /* Listen for changes to the place entry search */
      files.notify["active-search"].connect (
        (obj, pspec) => {
          var _results_model = files.entry_renderer_info.results_model;
          var _groups_model = files.entry_renderer_info.groups_model;
          update_search_async.begin (files.active_search,
                                     (Section)files.active_section,
                                     _results_model, _groups_model);
        }
      );

      /* Listen for changes to the global search */
      files.notify["active-global-search"].connect (
        (obj, pspec) => {
          var _results_model = files.global_renderer_info.results_model;
          var _groups_model = files.global_renderer_info.groups_model;
          update_search_async.begin (files.active_global_search,
                                     Section.ALL_FILES,
                                     _results_model, _groups_model);
        }
      );

      /* We should not do anything with the results modelresults = yield index.search (search.get_search_string (),
                                          new Zeitgeist.TimeRange.anytime(),
                                          templates,
                                          Zeitgeist.StorageState.ANY,
                                          100,
                                          Zeitgeist.ResultType.RELEVANCY,
                                          null);
       * until we receieve the 'ready' signal */
      results_model.ready.connect (this.on_results_model_ready);

      sections_model.connect ();
      groups_model.connect ();
      global_groups_model.connect ();
      results_model.connect ();
      global_results_model.connect ();

      /* The last thing we do is export the controller. Once that is up,
       * clients will expect the SharedModels to work */
      control = new Unity.Place.Controller ("/com/canonical/unity/filesplace");
      control.add_entry (files);
      try {
        control.export ();
      } catch (DBus.Error error) {
        critical ("Failed to export DBus service for '%s': %s",
                  control.dbus_path, error.message);
      }
    }

    private void populate_sections ()
    {
      var sections = files.sections_model;

      if (sections.get_n_rows() != 0)
        {
          critical ("The sections model should be empty before initial population");
          sections.clear ();
        }

      sections.append (SectionsColumn.DISPLAY_NAME, "All Files",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Documents",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Folders",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Images",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Videos",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Presentations",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Other",
                       SectionsColumn.ICON_HINT, "", -1);
    }

    private void populate_groups ()
    {
      var groups = files.entry_renderer_info.groups_model;

      if (groups.get_n_rows() != 0)
        {
          critical ("The groups model should be empty before initial population");
          groups.clear ();
        }

      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Top Results",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Today",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Yesterday",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "This week",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Last Week",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "This Month",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Past Six Months",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "This Year",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Last Year",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "Specific Year", // FIXME: Use actual year, eg "2009"
                     GroupsColumn.DISPLAY_NAME, "Today",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Today",
                     GroupsColumn.ICON_HINT, "", -1);

      months = new Gee.ArrayList<unowned Dee.ModelIter?>();
      months.add(null);
      for (uint i = 1; i <= DateMonth.DECEMBER; i++)
      {
        unowned Dee.ModelIter iter = groups.append (
                                  GroupsColumn.RENDERER, "UnityDefaultRenderer",
                                  GroupsColumn.DISPLAY_NAME, Utils.get_month_name ((DateMonth)i),
                                  GroupsColumn.ICON_HINT, "", -1);
        months.add(iter);

      }
    }

    private void prepare_section_templates ()
    {
      PtrArray templates;
      Event event;

      /* HACK ALERT: All the (event as GLib.Object).ref() are needed because
       *             GPtrArray doesn't grab a ref to the event objects */

      /* Section.ALL_FILES */
      templates = new PtrArray.sized(1);
      event = new Event.full("", ZG_USER_ACTIVITY, "",
                             new Subject.full ("file:*",
                                               "",
                                               NFO_FILE_DATA_OBJECT,
                                               "", "", "", ""));
      templates.add ((event as GLib.Object).ref());
      section_templates.add (templates);

      /* Section.DOCUMENTS
       * FIXME: Filter out presentations: https://bugs.launchpad.net/zeitgeist/+bug/592599 */
      templates = new PtrArray.sized(1);
      event = new Event.full("", ZG_USER_ACTIVITY, "",
                             new Subject.full ("file:*",
                                               NFO_DOCUMENT,
                                               NFO_FILE_DATA_OBJECT,
                                               "", "", "", ""));
      templates.add ((event as GLib.Object).ref());
      section_templates.add (templates);

      /* Section.FOLDERS.
       * FIXME: We probably need to be clever here and use something
       *       like subject.origin in stead of NFO_FOLDER */
      templates = new PtrArray.sized(1);
      event = new Event.full("", ZG_USER_ACTIVITY, "",
                             new Subject.full ("file:*",
                                               NFO_FOLDER,
                                               NFO_FILE_DATA_OBJECT,
                                               "", "", "", ""));
      templates.add ((event as GLib.Object).ref());
      section_templates.add (templates);

      /* Section.IMAGES */
      templates = new PtrArray.sized(1);
      event = new Event.full("", ZG_USER_ACTIVITY, "",
                             new Subject.full ("file:*",
                                               NFO_IMAGE,
                                               NFO_FILE_DATA_OBJECT,
                                               "", "", "", ""));
      templates.add ((event as GLib.Object).ref());
      section_templates.add (templates);

      /* Section.VIDEOS */
      templates = new PtrArray.sized(1);
      event = new Event.full("", ZG_USER_ACTIVITY, "",
                             new Subject.full ("file:*",
                                               NFO_VIDEO,
                                               NFO_FILE_DATA_OBJECT,
                                               "", "", "", ""));
      templates.add ((event as GLib.Object).ref());
      section_templates.add (templates);

      /* Section.PRESENTATIONS
       * FIXME: Zeitgeist logger needs to user finer granularity
       *        on classification as I am not sure it uses
       *        NFO_PRESENTATION yet */
      templates = new PtrArray.sized(1);
      event = new Event.full("", ZG_USER_ACTIVITY, "",
                             new Subject.full ("file:*",
                                               NFO_PRESENTATION,
                                               NFO_FILE_DATA_OBJECT,
                                               "", "", "", ""));
      templates.add ((event as GLib.Object).ref());
      section_templates.add (templates);

      /* Section.OTHER
       * FIXME: Zeitgeist doesn't support collated AND NOT queries yet */
      templates = new PtrArray.sized(1);
      event = new Event.full("", ZG_USER_ACTIVITY, "",
                             new Subject.full ("file:*",
                                               "",
                                               NFO_FILE_DATA_OBJECT,
                                               "", "", "", ""));
      templates.add ((event as GLib.Object).ref());
      section_templates.add (templates);
    }

    private bool search_is_invalid (Search? search)
    {
      return search == null ||
             search.get_search_string () == null ||
             search.get_search_string () == "";
    }

    public void on_results_model_ready (Dee.SharedModel model)
    {
      var section = (Section)files.active_section;
      var _results_model = files.entry_renderer_info.results_model;
      var _groups_model = files.entry_renderer_info.groups_model;
      update_without_search_async.begin(section, _results_model, _groups_model);
      monitor.events_inserted.connect (on_events_inserted);

      // FIXME: Use the section queries as monitor templates as well
      //        and make sure we de-dupe the existing results as updates
      //        trickle in.
      //        See https://bugs.launchpad.net/anjali/+bug/598078
      //log.install_monitor.begin (monitor, null);
    }

    private string prepare_search_string (Search? search)
    {
      var s = search.get_search_string ();

      if (s.has_suffix (" "))
        s = s.strip ();

      if (!s.has_suffix ("*"))
        s = s + "*";

      return s;
    }

    private async void update_search_async  (Search search,
                                             Section section,
                                             Dee.Model results_model,
                                             Dee.Model groups_model)
    {
      if (search_is_invalid (search))
        {
          update_without_search_async.begin (section,
                                             results_model,
                                             groups_model);
          return;
        }

      var search_string = prepare_search_string (search);
      var templates = section_templates.get((int)section);

      try {
        /* Get relevancy ranked results for the "Top Results" group */
        var results = yield index.search (search_string,
                                          new Zeitgeist.TimeRange.anytime(),
                                          templates,
                                          Zeitgeist.StorageState.ANY,
                                          100,
                                          Zeitgeist.ResultType.RELEVANCY,
                                          null);

        /* Don't clear the results before the first ones are ready */
        results_model.clear ();
        Unity.FilesPlace.append_top_results (results,
                                             results_model,
                                             groups_model);

        debug ("Found %u/%u matches for search '%s'",
               results.size (), results.estimated_matches (), search_string);

        results = yield index.search (search_string,
                                      new Zeitgeist.TimeRange.anytime(),
                                      templates,
                                      Zeitgeist.StorageState.ANY,
                                      100,
                                      Zeitgeist.ResultType.MOST_RECENT_SUBJECTS,
                                      null);
        Unity.FilesPlace.append_events_sorted (results,
                                               results_model, groups_model);

      } catch (GLib.Error e) {
        warning ("Error performing search '%s': %s",
                 search.get_search_string (), e.message);
      }
    }

    private async void update_without_search_async (Section section,
                                                    Dee.Model results_model,
                                                    Dee.Model groups_model)
    {
      var sections_model = files.sections_model;

      if (Section.LAST_SECTION != sections_model.get_n_rows())
        {
          critical ("Section model malformed");
          return;
        }

      if (section > Section.LAST_SECTION)
        {
          critical ("Active section out of bounds: %u", section);
          return;
        }

      var templates = section_templates.get((int)section);

      try {
        var events = yield log.find_events (
                                        new Zeitgeist.TimeRange.anytime(),
                                        templates,
                                        Zeitgeist.StorageState.ANY,
                                        100,
                                        Zeitgeist.ResultType.MOST_RECENT_SUBJECTS,
                                        null);
        debug ("Got %u events for section %u", events.size(), section);

        /* Don't clear the model before we have the results ready */
        results_model.clear();

        Unity.FilesPlace.append_events_sorted (events,
                                               results_model, groups_model);
      } catch (GLib.Error e) {
        warning ("Error fetching recetnly used files: %s", e.message);
      }
    }

    private void on_events_inserted (Zeitgeist.Monitor mon,
                                     Zeitgeist.TimeRange time_range,
                                     ResultSet events)
    {
      // FIXME: Since we don't really have the timestamps for the evens in
      //        our model, we can't insert the events in the correct place
      //        although it's likely fine to just prepend them

      var results_model = files.entry_renderer_info.results_model;
      var groups_model = files.entry_renderer_info.groups_model;

      foreach (var ev in events)
        {
          if (ev.num_subjects() > 0)
            {
              // FIXME: We only use the first subject...
              Zeitgeist.Subject su = ev.get_subject(0);

              string icon = Utils.get_icon_for_uri (su.get_uri (),
                                                    su.get_mimetype ());
              uint group_id = Utils.get_time_group (ev, groups_model);

              debug ("Notify %s, %s, %u", su.get_uri(), su.get_mimetype(), group_id);

              results_model.prepend (ResultsColumn.URI, su.get_uri(),
                                     ResultsColumn.ICON_HINT, icon,
                                     ResultsColumn.GROUP_ID, group_id,
                                     ResultsColumn.MIMETYPE, su.get_mimetype(),
                                     ResultsColumn.DISPLAY_NAME, su.get_text(),
                                     ResultsColumn.COMMENT, su.get_uri(),
                                     -1);
            }
        }
    }
  }

  /* Appends a set of Zeitgeist.Events to our Dee.Model assuming that
   * these events are already sorted with descending timestamps */
  public void append_events_sorted (ResultSet events,
                                    Dee.Model results,
                                    Dee.Model groups)
    {
      foreach (var ev in events)
        {
          if (ev.num_subjects() > 0)
            {
              // FIXME: We only use the first subject...
              Zeitgeist.Subject su = ev.get_subject(0);

              string icon = Utils.get_icon_for_uri (su.get_uri (),
                                                    su.get_mimetype ());
              uint group_id = Utils.get_time_group (ev, groups);

              debug ("Got %s, %s, %u, %s", su.get_uri(), su.get_mimetype(), group_id, icon);

              results.append (ResultsColumn.URI, su.get_uri(),
                              ResultsColumn.ICON_HINT, icon,
                              ResultsColumn.GROUP_ID, group_id,
                              ResultsColumn.MIMETYPE, su.get_mimetype(),
                              ResultsColumn.DISPLAY_NAME, su.get_text(),
                              ResultsColumn.COMMENT, su.get_uri(),
                              -1);
            }
        }
    }

  /* Appends a set of Zeitgeist.Events to our Dee.Model assuming that
   * these events are already sorted according to query relevancy.
   * The results will be added to the Top Results group */
  public void append_top_results (ResultSet events,
                                  Dee.Model results,
                                  Dee.Model groups)
    {
      Gee.Set<string> seen_uris = new Gee.HashSet<string>(str_hash);

      foreach (var ev in events)
        {
          if (ev.num_subjects() > 0)
            {
              // FIXME: We only use the first subject...
              Zeitgeist.Subject su = ev.get_subject(0);

              /* De-dup the results keyed on the subject URIs */
              var uri = su.get_uri();
              if (uri in seen_uris)
                continue;
              seen_uris.add (uri);

              string icon = "";

              debug ("Got %s, %s, %u", uri, su.get_mimetype(), Group.TOP_RESULTS);

              results.append (ResultsColumn.URI, uri,
                              ResultsColumn.ICON_HINT, icon,
                              ResultsColumn.GROUP_ID, Group.TOP_RESULTS,
                              ResultsColumn.MIMETYPE, su.get_mimetype(),
                              ResultsColumn.DISPLAY_NAME, su.get_text(),
                              ResultsColumn.COMMENT, uri,
                              -1);
            }
        }
    }

} /* namespace */
