/*
 * Copyright (C) 2010 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Mikkel Kamstrup Erlandsen <mikkel.kamstrup@canonical.com>
 *
 */

using Zeitgeist;

namespace Unity.FilesPlace.Utils {

    const string icon_attribs = FILE_ATTRIBUTE_PREVIEW_ICON + "," +
                                FILE_ATTRIBUTE_STANDARD_ICON + "," +
                                FILE_ATTRIBUTE_THUMBNAIL_PATH;

	/* Synchronous method to query GIO for a good icon for a uri/mimetype */
    public string get_icon_for_uri (string uri, string mimetype)
    {
      try {

        File f = File.new_for_uri (uri);
        FileInfo info = f.query_info (icon_attribs,
                                      FileQueryInfoFlags.NONE,
                                      null);
        string thumbnail_path = info.get_attribute_byte_string (FILE_ATTRIBUTE_THUMBNAIL_PATH);

        if (thumbnail_path == null)
          {
            Icon icon = info.get_icon();
            return icon.to_string();
          }
        else
          return new FileIcon(File.new_for_path (thumbnail_path)).to_string ();
      } catch (GLib.Error e) {
        /* We failed to probe the icon. Try looking up by mimetype instead */
        Icon icon2 = g_content_type_get_icon (mimetype);
        return icon2.to_string ();
      }
    }

    /* Async method to query GIO for a good icon for a uri/mimetype */
    public async string get_icon_for_uri_async (string uri, string mimetype)
    {
      try {

        File f = File.new_for_uri (uri);
        FileInfo info = yield f.query_info_async (icon_attribs,
                                                  FileQueryInfoFlags.NONE,
                                                  Priority.LOW,
                                                  null);
        Icon icon = info.get_icon();
        string thumbnail_path = info.get_attribute_byte_string (FILE_ATTRIBUTE_THUMBNAIL_PATH);

        if (thumbnail_path == null)
          return icon.to_string();
        else
          return thumbnail_path;
      } catch (GLib.Error e) {
        /* We failed to probe the icon. Try looking up by mimetype instead */
        Icon icon2 = g_content_type_get_icon (mimetype);
        return icon2.to_string ();
      }
    }

  public string get_month_name (DateMonth month)
    {
      switch (month)
        {
          case DateMonth.BAD_MONTH:
            return "Invalid Month";
          case DateMonth.JANUARY:
            return "January";
          case DateMonth.FEBRUARY:
            return "February";
          case DateMonth.MARCH:
            return "March";
          case DateMonth.APRIL:
            return "April";
          case DateMonth.MAY:
            return "May";
          case DateMonth.JUNE:
            return "June";
          case DateMonth.JULY:
            return "July";
          case DateMonth.AUGUST:
            return "August";
          case DateMonth.SEPTEMBER:
            return "September";
          case DateMonth.OCTOBER:
            return "October";
          case DateMonth.NOVEMBER:
            return "November";
          case DateMonth.DECEMBER:
            return "December";
        }
        
      return "Internal Error";
    }
    
    /* Calculate the time group of the event returning the group id
     * of the group @event belongs to. It may be necessary for this method
     * to add a row to @groups_model if the group is one of the variadic
     * types such as */
    public uint get_time_group (Zeitgeist.Event event, Dee.Model groups_model)
    {
      // FIXME: We need to calculate bounds of previous midnight, week, month etc
      //        This impl is just a quick hack to have something to show
      
      var t = event.get_timestamp ();
      var now = Timestamp.now ();
      
      /* Set limit to the previous midnight */
      var limit = Timestamp.prev_midnight (now);
      if (t > limit)
        return Group.TODAY;
      else if (t > limit - Zeitgeist.Timestamp.DAY)
        return Group.YESTERDAY;
      
      /* Set limit to start of calendar week (monday) */
      var date = Date();
      Timestamp.to_date (now, out date);
      limit = limit - (date.get_weekday() - 1)*Timestamp.DAY;
      if (t > limit)
        return Group.THIS_WEEK;
      else if (t > limit - Zeitgeist.Timestamp.WEEK)
        return Group.LAST_WEEK;
      
      /* Set limit to start of current calendar month */
      date.set_day (1);
      limit = Timestamp.from_date (date);
      if (t > limit)
        return Group.THIS_MONTH;
      
      /* Set limit to start of previous calendar month */
      date.subtract_months (1);
      limit = Timestamp.from_date (date);
      if (t > now - 6*30*Zeitgeist.Timestamp.DAY)
        return Group.PAST_SIX_MONTHS;
      
      /* Set limit to 1st pf January in the current year. We need to 
       * reset 'date' because date.subtract_months(1) above might have
       * turned a year */
      Timestamp.to_date (now, out date);
      date.set_month (DateMonth.JANUARY);
      date.set_day (1);
      limit = Timestamp.from_date (date);
      if (t > limit)
        return Group.THIS_YEAR;
      else
        return Group.LAST_YEAR;
    }
  
} /* namespace */
